#include <math.h>
#include <stdio.h>
#include <malloc.h>
/*  #include <iostream.h> only needed when main is activated
*/

#include "general.h"
#include "matrix.h"
#include "alloc.h"

/**********************************************************************/
/* Version : April 10, 1996 */
#define TINY 1.0e-20

void ludcmp(real **a, int n, int *indx, real *d)
/*
Given a matrix a[1..n][1..n], this routine replaces it by the LU decomposition 
of a rowwise permutation of itself.  a and n are input.  a is output, 
indx[1..n] is an output vector that records the row permutation effected 
by the partial pivoting; d is output at +-1 depending on whether the 
number of row interchanges was even or odd, respectively.  This routine is 
used in cominbation with lubksb to solve linear equations or invert a matrix
*/
{
	int i,imax,j,k;	
	real big,dum,sum,temp;
	real *vv = NULL;
	
        alloc(n,&vv);
	*d=1.0;
	for (i=0;i<=n-1;i++) {
		big=0.0;
		for (j=0;j<=n-1;j++)
			if ((temp=fabs(a[i][j]))>big) big= temp;
		if (big ==0.0) 
                  nrerror("inverse.c : Singular matrix in routine ludcmp");
		vv[i]=1.0/big;
	}
	for (j=0;j<=n-1;j++) {
		for (i=0;i<j;i++) {
		 	sum=a[i][j];
			for (k=0;k<i;k++) sum -= a[i][k]*a[k][j];
			a[i][j] = sum;
		}
		big = 0.0; 
		for (i=j;i<=n-1;i++) {
			sum=a[i][j];
			for (k=0;k<j;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j] = sum;
			if ((dum=vv[i]*fabs(sum)) >=big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) {
			for (k=0;k<=n-1;k++) {
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			*d = -(*d);
			vv[imax]=vv[j];
		}
		indx[j] = imax;
		if (a[j][j] == 0.0) a[j][j]=TINY;
		if (j != n-1) {
			dum=1.0/(a[j][j]);
			for (i=j+1;i<=n-1;i++) a[i][j] *= dum;
		}
	}
	frei(n,&vv);
}	
#undef TINY

/**********************************************************************/
/* Version : April 10, 1996 */

/*
Solves the set of n linear equations AX=B.  Here a[1..n][1..n] is input, 
not as the matrix A but rather as its LU decomposition, determined 
by the routine ludcmp.  indx[1..n] is input as the permutation vector 
returned by ludcmp.  b[1..n] is input as the right-hand side vector
B, and returns with the solution vector X. a, n and indx are not modified 
by this routine and can be left in place for successive calls with 
different right-hand sides b.  This routine takes into account the 
possibility that b will begin with many zero elements, so it is efficient
for use in matrix inversion.
*/
void lubksb(real **a, int n, int *indx, real b[])
{
	int i,ii= -1,ip,j;
	real sum;

	for (i=0;i<=n-1;i++) {
		ip = indx[i];
		sum = b[ip];
		b[ip] = b[i];
		if (ii != -1)
			for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
		else if (sum) ii=i;
		b[i] = sum;
	}
	for (i=n-1;i>=0;i--) {
		sum=b[i];
		for (j=i+1;j<=n-1;j++) sum -= a[i][j]*b[j];
		b[i]=sum/a[i][i];
	}
}

/**********************************************************************/
/* Version : May 19, 1996 */
/* calculates the inverse of a matrix */
/* y contains the inverse, 
   a will have been DESTROYED (so copy before if still needed) 
*/
void inverse(real **a, real **y, int N)
{
	void ludcmp(real **a, int n, int *indx, real *d);
	void lubksb(real **a, int n, int *indx, real b[]);
	int i,j, *indx=NULL;
	real d, *col=NULL;
	
	alloc(N,&indx);
	alloc(N,&col);
	ludcmp(a,N,indx,&d);
	for (j=0;j<=N-1;j++) 
 	{
		for (i=0;i<=N-1;i++) col[i]=0;
		col[j]=1.0;
		lubksb(a,N,indx,col);
		for (i=0;i<=N-1;i++) y[i][j]=col[i];
	}
        frei(N,&indx);
        frei(N,&col);
}

/**********************************************************************/
/* Version : April 10, 1996 */
/* calculates the determinant of a matrix */
real deter(real **a, int N)
{
	void ludcmp(real **a, int n, int *indx, real *d);
	real d;
	int j, *indx=NULL;

	alloc(N,&indx);
	ludcmp(a,N,indx,&d);
	for (j=0; j<N;j++) d *= a[j][j];
        frei(N,&indx);
	return d;
}
/**********************************************************************/
/*
void main(void)
{
	int size,i,j;
	real **myarray=NULL,**copy=NULL;
	real **myarrayinv=NULL;
	real det;

cout << myarray << space  << copy << space << myarrayinv << endl;
	size=3;
    print_list("int_list");
    print_dlist("int_list");
	alloc(size,size,&myarray);
cout << myarray << space  << copy << space << myarrayinv << endl;
    print_list("int_list");
    print_dlist("int_list");
	alloc(size,size,&copy);
cout << myarray << space  << copy << space << myarrayinv << endl;
    print_list("int_list");
    print_dlist("int_list");
	alloc(size,size,&myarrayinv);
cout << myarray << space  << copy << space << myarrayinv << endl;
    print_list("int_list");
    print_dlist("int_list");
	myarray[0][0] = 1.0; myarray[0][1] = 2.0; myarray[0][2] = 3.0;
	myarray[1][0] = 1.0; myarray[1][1] = 4.0; myarray[1][2] = 7.0;
	myarray[2][0] = 1.0; myarray[2][1] = 9.0; myarray[2][2] = 10.0; 
	for (i=0;i<size;i++)
		for (j=0;j<size;j++)
			copy[i][j] = myarray[i][j];
	inverse(myarray,myarrayinv,size);
        printmatr(size,size,myarrayinv,"Inverse:");
	det = deter(copy,size);
	printf("The determinant is %f \n",det);
}
*/
