/***************************/
/* Kriging model functions */
/***************************/

#ifndef general
#include "general.h"
#endif

#ifndef  _KRIG_
#define  _KRIG_

#define NoCV  -1   /*No cross validation */

struct InitStruct
{
  boolean X;
  boolean y;
  boolean R;
  boolean F;
  boolean beta;
  boolean alpha;
  boolean theta;
  boolean U;
  boolean var;   //errorvar + stochvar
  boolean range;
  boolean d;
  boolean n;
};

struct KrigArray
{
  boolean init; // already initialized ?
  boolean ydescrip_init; // already initialized ?
  int p;     //number of y responses
  int c ;    //number of y constraints
  int *pvec; //vec indicating the position of responses within Krigarray 
  int *cvec; //vec indicating the position of constraints within Krigarray 
  real **ydescrip; // ydescription matrix
};

/* everything needs to be a pointer (using *'s) because otherwise 
   changes in function are not passed back to the calling function 
  (unless we are always passing a pointer to the structure) */ 
struct KrigModel
{
  int *n;  	/* scalar,  number of obs */
  int *d;  	/* scalar, dimension of x */
  int *r; 	/* scalar, dimension of regression model */
  real *var;    /* scalar, total variance */
  real *stochratio;  // scalar, stochvar/(errorvar+stochvar)= stochvar/var
  real *beta; 	/* r,1 vector */
  real *alpha;  /* d,1 vector */
  real *theta; 	/* d,1 vector*/
  real *y;  	/* n,1 vector */
  real *eigval; /* n+r,1 vector, eigenvalues of augmented R */
  real *eigvalinv; /* n+r,1 vector, inverse eigenvalues of augmented R */
  real **F;	/* n,r matrix, (constant) regression matrix */
  real **X;  	/* n,d matrix */
  real **R; 	/* n,n matrix */
  real **U;     /* n+r,n+r matrix, normalized eigenvectors of augmented R,
                   corresp. to eigval */
  real **range; /* d,2 matrix for ranges of x variables */ 
  string *xlabel; // d,1 label for x
  string ylabel; // one string only, label for y
  InitStruct init[1];
  
  real *loglike;/* scalar, max. loglikelihood */
};


/***************************/
/* ... read matrices ... */
/***************************/
int read_corpar(KrigModel *mine,KrigArray *ka,char *filename);
int read_x(KrigModel *kr,char *filename);
int read_beta(KrigModel **ptrkr,KrigArray *ka,char *filename);
int read_y(KrigModel **ptrkr,KrigArray *ka,char *filename);
int read_ydescrip(KrigModel **kr,KrigArray *ka,char *filename);
int read_xdescrip(KrigModel *kr,KrigArray *ka,char *filename);
int read_matrix(int oldn, int oldm,string ***mat1, int *newn, int *newm,
   char *filename, string **labelvec1);
int read_matrix(int oldn, int oldm,real ***mat1, int *newn, int *newm,
   char *filename, string **labelvec1);
/***************************/
/* ... write matrices ... */
/***************************/
void outputmatr(int n, int m, real **matrix,char matname[],
     string varnames[], char varnames2[], char comment[],
     char mode='w', boolean exact=FALSE);
void outputmats(int n, int m,string **matrix,char matname[],
     string varnames[], char varnames2[], char comment[],
     char mode='w', boolean exact=FALSE);
void outputvecr(int n, real *vector,char vecname[],
     char varnames[], char comment[],char mode='w');
void output_range(KrigModel *kr, char name[],char *comment);
void outputmatr_cover(int n_row,int n_col,real **matrix,char filename[],
   string *label, string comment);
void outputmats_cover(int n_row,int n_col,string **matrix,char filename[],
   string *label, string comment);
/***************************/
/* ... kriglib : auxiliary Kriging routines ... */
/***************************/
void kr_calcR(KrigModel *kr);
void calcR(real **X,real *alpha,real *theta, real stochratio,
   int n,int d,real **R);
void calcr(real *newx, real *r,real **X,real *alpha,real *theta,
   real stochratio, int n,int d);
real calcRij(KrigModel *kr, real *vec1, real *vec2);
void kr_calcF(KrigModel *kr);
void calcF(int n,int r,real **F);
void kr_calc_augmR(KrigModel *kr, int cvindex=NoCV);
void kr_cover_augmR(KrigModel *kr);
void calc_augmR(int n,int r,real **R,real **F,real **U,real *ev,real *evinv);
int test_X_in_range(int n,int d,real **mat,real **range);
int check_constraints(int n,int d,real **mat,real **simp_const);
int construct_ymat(KrigModel *kr, KrigArray *ka,real **ymat,char *ylab, int maxylab);
void construct_cormat(KrigModel *kr,KrigArray *ka,string **cormat,
  string *corlab);
int construct_ydescrip(KrigArray *ka,KrigModel *kr,string ***M, string **label, int *pc, int *n_label);
int construct_beta(KrigArray *ka, KrigModel *kr,string ***addrM, string **addrlabel, int *pc,int *n_label);
void label_suffix(KrigModel *kr,int pc,int size_set,string *label);
void copy_vec_corpar(KrigModel *kr, real *parvec, int  n_params);
void assemble_ylab(KrigArray *ka,KrigModel *kr,string *ylab);
void print_corpars(KrigModel *kr);
/***************************/
/*... kriglib: Initialization checks/safeguards... */
/***************************/
boolean initialized(InitStruct *init, char *n1, char *n2,
  char *n3, char *n4, char *n5,boolean output=TRUE);
boolean initialized(KrigArray *ka);
void ensure_init(KrigModel *addr_kr, char name[]);
void equal_init(KrigModel *kr,char *command);
void database(KrigModel *kr, KrigArray *ka);
void info(int n_token);
int tokenize(char *l,int *n_token,int len_token,char **token);
void fprint(char *text);
void output_error(KrigModel *kr,char error_text[]);
void output_temp(KrigModel *kr,char *top_text);
/***************************/
/* ... fitting .... */
/***************************/
void calc_stochvar(real **U,real *diaginv);
void calc_beta(real **U,real *diaginv);
real krigloglike(real *parvec, int d);
void krig_mle(KrigModel *kr,boolean details);
/***************************/
/* ... BLUP, MSE , COV, Cross Validation ... */
/***************************/
void krigcov(KrigModel *kr, int npred, real **xpred, real **covmat, boolean newpt);
void krigblup(KrigModel *kr, int npred, real **xpred, real **ypred, 
     int cvindex=NoCV);
void krigpredict(KrigModel *kr,KrigArray *ka, string inputfile,
  string outputfile);
void krigcv(KrigModel *kr,real **ypred,int col_ypred);
void krigcv_anyXY(KrigModel *kr,real **anyX,real *anyY,int an,int ad,
   real **cvmat,int col_cvmat);
real variance_check(KrigModel *kr, real mse, int counter);
real krigmse_anyX(KrigModel *kr, real **anyX,int n_rows,int n_cols,int npred,
 real **xpred);
void krigblup_anyXY(KrigModel *kr,real **anyX,real *anyY, int an,int ad,
     int npred,real **xpred, real **ypred);
void free_krigblup_anyXY(int an, int r);
void init_krigblup_anyXY(KrigModel *kr, real **anyX,int an, int ad, int r);
/***************************/
/* ... Optimization...     */
/***************************/
void sample_new_y(char FuncName[MAXLINE],int d,real *xbest, real ny,real *newy);
real ExpectedImprovement(real *x, int d);
int MaximalExpectedImprovement(KrigModel *kr,real *MaxExpImp,real *xbest);
void calc_ymin(KrigArray *ka,KrigModel *kr);
int minimize(KrigModel *kr,KrigArray *ka);
int MinimizeStage(KrigModel *kr, KrigArray *ka,int number, char *outputfile);
boolean ConvergeCrit1(real y,real deltay,real TolAbs,real TolRel);
boolean ConvergeCrit(real y,real deltay);
real recurse_integral(int p,real pdfprob, real cdfprob, real zmin);
/***************************/
/* ... Effects ... 	   */
/***************************/
void MainEffects(KrigModel *kr,char *outfile, char mode='w');
void JointEffects(KrigModel *kr,char *outfile, char mode='w');
/***************************/
/* ... 1st order statistic...*/
/***************************/
int orderalg(KrigModel *kr, int N, real *mean, real **xpred, real *bigh,
  int npred);
/***************************/
/*... krigmodel functions ... */
/***************************/
void init_KrigModel(KrigModel *kr, int n , int d, int r);
void init_KrigArray(KrigArray *ka);
void default_KrigArray(int oldn,KrigArray *ka,int n);
void undo_default_KrigArray(KrigArray *ka,int n);
void adjust_KrigArray(KrigArray *ka);
void copy_KrigModel(KrigModel *kr1,KrigModel *kr2);
void NULL_KrigModel(KrigModel *kr);
void NULL_KrigArray(KrigArray *ka);
int change_krig_p(KrigArray *ka,KrigModel **ptrkr,int newm);
void print_krigmodel(KrigModel *kr);
void free_KrigModel(KrigModel *kr);
void free_KrigArray(KrigArray *ka, int n);
void free_Krig_cover(KrigArray *ka,KrigModel *kr);
/***************************/
/*... krigmain functions ... */
/***************************/
void menu_print(KrigModel **ptrkr,KrigArray *ka,int action,char *p,char *filename,char *top_text="hallo");
void inout(KrigModel **ptrkr,KrigArray *ka,char **token,int n_token);
void equal_init(KrigModel *kr,char **token,int n_token);
void menu(KrigModel **ptrkr,KrigArray *ka);
/***************************/
/*... krigcover functions ... */
/***************************/
void init_MinStage(KrigModel *kr,KrigArray *ka,char **token,int n_token);
void init_Minimize(KrigModel *kr,KrigArray *ka,char **token,int n_token,
   boolean RandomError);
void init_MainEffects(KrigModel *kr,KrigArray *ka,char **token,int n_token);
void init_JointEffects(KrigModel *kr,KrigArray *ka,char **token,int n_token);
void init_predict(KrigModel *kr,KrigArray *ka,char **token, int n_token);
void init_fit(KrigModel *kr,KrigArray *ka,char **token,int n_token);
void init_krigcv(KrigModel *kr,KrigArray *ka,char **token,int n_token);
/***************************/
/* ... krprint to both afp and stdout... */
/***************************/
void fprint(char *text);
#endif
