#include <stdio.h>
#include <math.h>

#include "general.h"
#include "alloc.h"
#include "matrix.h"
#include "krig.h"

#define DEBUG 0

extern boolean RandomError;
extern FILE *ofp;

// for communication with mse_anyX
static real **aU=NULL, *a_eigv=NULL, *a_eigvinv=NULL, **aR=NULL, **aF;
static boolean mse_anyX_flag=FALSE; //TRUE if initialized , FALSE otherwise

/************************************************************************/
/* Version : Jan 25, 1997 						*/
/* Purpose : checks whether variance is negative and , pending on       */
/*	whether the value is deemed to be numerically 0 or actually 	*/
/* 	wrong an error message is issued / var set to 0			*/  
/* on input : mse    variance						*/
/*	counter     number of the observation, ranging from 0..n	*/
/* NOTE  : procedure only needs to be called is var <0.0		*/
/* on return : var   (either old value or 0.0)				*/
real variance_check(KrigModel *kr, real mse, int counter) 
{
     if (mse < -.1)
     {//care: if y's range over several orders of magnitude, errors can be large
        fprintf(ofp,"Observation Number  %d\n",counter+1);
        fprintf(ofp,"mse : %f\n");
        output_error(kr,"krigblup,krigcov or mse_anyX: MSE in covariance is negative, possibly due to scaling.");
        //nrerror("krigblup: MSE is negative");
     }
     else if (mse< -1.0e-7)
     {
       fprintf(ofp,"WARNING : MSE for obs. %d is %f.Set to zero.\n", 
           counter+1,mse);
       mse = 0.0;
     }
     else if (mse < 0 ) mse=0.0;
     return mse;
}
/************************************************************************/
/* Version : July 25, 1996 						*/
/* Kriging covariance of points (MSE is also calculated in BLUP) 	*/
/* input :  npred :  dimension of xpred					*/
/*	    xpred :  npred,npred x prediction matrix			*/
/*	    kr    :  Kriging Model					*/
/*	    newpt :  FALSE : if only dealing with sample 		*/
/*	             TRUE  : if one or more pts are new, changes calc R12*/ 
/* Output : covmat contains covariance matrix 				*/
/* Note: could be made more efficient by calulating  first a vector of  */ 
/*       vec_i=(f_i,r_i) for i=1..npred rather than doing it one by one */
/* Note : U, eigval,eigvalinv have to be calc before 			*/
/* Note : the covariances calculated differ by about .01% from will's	*/
/* NOTE : this is not tested with RANDOMERROR				*/
/*	in particular see line with cov= below				*/
/*      seems to be relavant only for order stats right now 		*/
void krigcov(KrigModel *kr, int npred, real **xpred, real **covmat, 
   boolean newpt)
{
  int i, n, r, d, counter1, counter2;
  real cov, s, R_x1x2;
  real *vec1=NULL, *vec2=NULL;
  real *r1=NULL, *r2=NULL;
  real *temp1=NULL, *temp2=NULL;

  if (newpt && RandomError) 
     nrerror("Code Bug : want to calc. covariance in the presence of random error.  \n Need to modify calcRij."); 
  
  n = *kr->n ;
  r = *kr->r ;
  d = *kr->d ;
  alloc(r+n,&vec1); alloc(r+n,&vec2);   
  alloc(n,&r1);   alloc(n,&r2);   
  alloc(r+n,&temp1); alloc(r+n,&temp2);

 for (counter1=0; counter1<npred; counter1++)
 {
  for (counter2=counter1; counter2<npred; counter2++)
  {
     /* construct vec1=(f1,r1)' and vec2=(f2,r2)' */
     for (i=0; i<r ; i++)
     {
       if (r==1) 
       { 
         vec1[0]= 1.0; 
         vec2[0]= 1.0;
       }
       else 
         nrerror("krigcov(): regression model not dealt with, r>1 "); 
     }

     calcr(xpred[counter1], r1,kr->X,kr->alpha,kr->theta,
         *kr->stochratio,*kr->n,*kr->d);
     calcr(xpred[counter2], r2,kr->X,kr->alpha,kr->theta,
         *kr->stochratio,*kr->n,*kr->d);
     //printvecr(d,xpred[counter1],"x1:"); printvecr(d,xpred[counter2],"x2:");
     //printvecr(n,r1,"covr1:"); printvecr(n,r2,"covr2:");
     for (i=0; i<n; i++)
     {
       vec1[i+r]= r1[i];
       vec2[i+r]= r2[i];
     }
  
     /* multiply s    = (vec1' U) Diag(eigval)^{-1}  (vec2' U)' */
     /*               = temp1' *   Diag(eigval)^{-1}  temp2  */
     /*               = temp1' *   temp2(overwritten)  */
     VecMultMatReal(vec1,n+r,kr->U,n+r,temp1); 
     VecMultMatReal(vec2,n+r,kr->U,n+r,temp2); 
     VecEwMultVecReal(kr->eigvalinv,n+r,temp2);
     s=VecMultVecReal(temp1,n+r,temp2);
     
     /* compute covariance for one pair */
     R_x1x2 = calcRij(kr,xpred[counter1],xpred[counter2]);
     cov = *kr->var * (R_x1x2 - s);
     covmat[counter1][counter2] = cov;
     covmat[counter2][counter1] = cov;  /* by symmetry of the cov matrix */
     if (counter2==counter1 && cov <0.0) 
        cov=variance_check(kr,cov,counter1);
  }
 }
 frei(r+n,&vec1); frei(r+n,&vec2);
 frei(n,&r1); frei(n,&r2);
 frei(r+n,&temp1); frei(r+n,&temp2);
}
/************************************************************************/
/* Version : July 15, 1996 						*/
/* Purpose: Blup and MSE (also for Cross Validation) 			*/
/* Missing : regression model 						*/
/* on input : npred  : number of obs to be predicted  			*/
/*            xpred  : npred,d matrix of prediction locations 		*/  
/* 	      cvindex: -1 (=NoCV) if no cross validation		*/
/*                   i otherwise (indicates which position to leave out)*/
/* on output: ypred  : npred,2 matrix of prediction and se's   		*/
/* Note : U, eigval,eigvalinv have to be calc before 			*/
/*        (local optimization uses this a gazillion times) 		*/
void krigblup(KrigModel *kr, int npred, real **xpred, real **ypred,int cvindex)
{
  int i, wn, r, counter; 
  int adjust;  /* adjust index in case of cross validation */
  real mse, blup, s ;
  real *vec1=NULL, *yvec=NULL;
  real *r1=NULL;
  real *temp1=NULL, *temp2=NULL;

  wn = *kr->n ;  /* working n , could be (n) or (n-1) */
  if (cvindex != -1)  wn--;   
  r = *kr->r ;

  if (DEBUG) printmatr(wn+r, wn+r, kr->U,"krigblup U");  
 
  alloc(r+wn,&vec1); alloc(r+wn,&yvec);   
  alloc(r+wn,&temp1); alloc(r+wn,&temp2);
  alloc(*kr->n,&r1);  /* need full r vector */   

  for (counter=0; counter<npred; counter++)
  {
     /* construct first half of  vec1=(f1,r1)' and yvec=(0,y)' */
     for (i=0; i<r ; i++)
     {
       if (r==1) 
       {	
         vec1[0]= 1.0;
         yvec[0]= 0.0;
       } 
       else 
         nrerror("krigblup(): regression model not dealt with, r>1 "); 
     }
     /*printmatr(1,2,ypred,"ypred1"); */

     calcr(xpred[counter], r1,kr->X,kr->alpha,kr->theta,
        *kr->stochratio,*kr->n,*kr->d);
     /* printvecr(d,xpred[counter],"x1:");  printvecr(*kr->n,r1,"r1:"); */ 
     /* construct second half of  vec1=(f1,r1)' and yvec=(0,y)' */
     for (adjust=0,i=0; i< *kr->n; i++)
     {
       /* note: if cvindex=-1 , adjust is always 0 */
       /* otherwise the (cvindex)th observation is skipped */
       if (i == cvindex)
         adjust = -1;
       else
       {
         vec1[i+r+adjust]= r1[i];
         yvec[i+r+adjust]= kr->y[i];
       }
     }
     if (DEBUG) {printvecr(wn+r,vec1,"vec1:"); printvecr(wn+r,yvec,"yvec:"); } 
  
     /* compute temp1 = (vec1' U) Diag(eigval)^{-1} */
     VecMultMatReal(vec1,wn+r,kr->U,wn+r,temp1); /*need for both MSE and BLUP*/
     VecEwMultVecReal(kr->eigvalinv,wn+r,temp1); /*need for both MSE and BLUP*/

     /* compute BLUP */
     /* multiply blup  = (vec1' U) Diag(eigval)^{-1}* (yvec' U)' */
     /* multiply       = temp1'                     *  temp2     */
     VecMultMatReal(yvec,wn+r,kr->U,wn+r,temp2);
     blup=VecMultVecReal(temp1,wn+r,temp2);
     ypred[counter][0] = blup;

     /* compute MSE */
     /* multiply s    = (vec1' U) Diag(eigval)^{-1}*  (vec1' U)' */
     /*               = temp1'                     *     temp2  */
     VecMultMatReal(vec1,wn+r,kr->U,wn+r,temp2); 
     s=VecMultVecReal(temp1,wn+r,temp2);
     mse =  *kr->var * (1.0 -  s);
     if (mse<0.0) 
        mse = variance_check(kr,mse,counter);
     ypred[counter][1] = sqrt(mse);

  }
  frei(r+wn, &vec1); frei(r+wn,&yvec);
  frei(*kr->n,&r1); 
  frei(r+wn,&temp1); frei(r+wn,&temp2);
}
/************************************************************************/
/* Version : July 10, 1996                                              */
/* Purpose : calc. Cross Validated E(I) and E(I^p) 			*/
/* input : matrix  (n,3=d) matrix, 1st col: blup, 2nd col: se		*/
/*         ymin    min(y)						*/
/* output : matrix 3rd col: CVEI 					*/
/* Note   : in a former version also CVEI^5 and CVEI^10 were calculated */
void cvei(real **matrix, real n, real d, real ymin)
{
  int k,i,j,globalp;
  int globalpvec[1];
  //former : int globalpvec[3]={1,5,10};
  real ei, pdfprob,cdfprob,z, mu , sigma;

  globalpvec[0]=1; // if change this , change col_ypred=5
  for (k=0; k<n;k++) 
  {
   mu = matrix[k][0];
   sigma = matrix[k][1];
   for (j=0;j<1;j++)
   { // for each global parameter in turn , formerly j<3
     globalp = globalpvec[j];
     if (sigma!=0.0)
     {
       z  = (ymin-mu) / sigma ;
       if (globalp == 1)
         ei = sigma*( z* cdfgauss(z,TRUE) + pdfgauss(z));
       else if (globalp>1)
       {
         pdfprob = pdfgauss(z);
         cdfprob = cdfgauss(z,TRUE);
         for (i=0, ei=0.0;i<=globalp;i++)
         {
           ei += choose(globalp,i) * pow(-1.0,real(i%2)) * pow(z,globalp-i)*
               recurse_integral(i,pdfprob,cdfprob,z);
         }
         // for numerical reasons calc sigma *  (...)^(1/p)
         // rather than               (sigma^p * (...))^(1/p)
         if (ei>1e-30)
         { // the number 1e-30 is guess work, adjust if necessary
           ei = pow(ei,1.0/real(globalp));//take globalp.th root, monotone trafo
           ei *= sigma;
         }
         else ei = 0.0;
       }
       else
         nrerror("ExpectedImprovement: globalp<1 ");
     }
     else ei = 0.0 ;
     matrix[k][2+j]= ei;
   }
  }
}
/************************************************************************/
/* Version : July 10, 1996                                              */
/* Purpose : Cross Validate the Kriging Model 				*/
/* Note    : F and R have to be calculated before 			*/
/* NOTE    : U, eigval , eigvalinv are reallocated and updated 		*/
void krigcv(KrigModel *kr,real **ypred,int col_ypred)
{
  void cvei(real **matrix, real n, real d, real ymin);
  int i,n, nr; 
  real *cvx=NULL; /* i^th obs to be cross validated */ 
  real ymin;

  n = *kr->n;
  nr = n + *kr->r;

  if (initialized(kr->init,"X","y","alpha","theta","beta") 
   && initialized(kr->init,"var",NULL,NULL,NULL,NULL)  ) 
  {
     ensure_init(kr,"R");
     ensure_init(kr,"F");
     morealloc(nr,nr,&kr->U,nr-1,nr-1);
     morealloc(nr,&kr->eigval,nr-1);
     morealloc(nr,&kr->eigvalinv,nr-1);
  
     /* cross validate at the i^th point denoted by cvx */
     for (i=0;i<n;i++)
     {
       cvx = kr->X[i];    /* care, cross validation x-vector  */
       kr_calc_augmR(kr,i);
       /* note the vectors r and y are adjusted within krigblup */ 
       krigblup(kr, 1, &cvx, &(ypred[i]), i);//ypred in krigblup is 1,2 matrix 
     }
     ymin = Real_Minimum(n,kr->y);//ignoring possible constraints
     cvei(ypred,n,col_ypred,ymin);

     /* for further use U, eigval, eigvalinv need to have proper dimensions */
     morealloc(nr-1,nr-1,&kr->U,nr,nr);
     morealloc(nr-1,&kr->eigval,nr);
     morealloc(nr-1,&kr->eigvalinv,nr);
     kr_cover_augmR(kr); //kr_calc_augmR(kr,NoCV); 
  }
}
/************************************************************************/
/* Version : July 10, 1996                                              */
/* Purpose : Cross Validate the Kriging Model for any X,Y (given corpars)*/
/* on input : anyX : an,ad x-matrix to be used for cross validation     */
/*	      anyY : ad,1  y vector correspnding to the X matrixa	*/	
/* 	      cvmat: an,col_cvmat matrix 				*/
/* on output: cvmat: 		contains BLUP,MSE,CVEI,..		*/
void krigcv_anyXY(KrigModel *kr,real **anyX,real *anyY,int an,int ad,  
   real **cvmat,int col_cvmat)
{
  void cvei(real **matrix, real n, real d, real ymin);
  int i,r; 
  real ymin; 
  real **anycvX, *anycvY;

  if (initialized(kr->init,"alpha","theta","beta","var",NULL)) 
  {
     r = *kr->r;
     alloc(an-1,ad,&anycvX);
     alloc(an-1,&anycvY);
     for (i=0;i<an;i++)
     { // cross validate at the i^th point denoted by anyX[i] 
       copy_real_matrix_minus_1row(an,ad,anyX,i,anycvX);
       copy_real_vector_minus_one(an,anyY,i,anycvY);
 //frei(an-1,ad,&anycvX); frei(an,ad,&anyX); exit(1);
       init_krigblup_anyXY(kr,anycvX,an-1,ad,r);
       krigblup_anyXY(kr,anycvX,anycvY,an-1,ad,1, &(anyX[i]), &(cvmat[i]));
       free_krigblup_anyXY(an-1,r);
     }
     ymin = Real_Minimum(*kr->n,kr->y);
         //ignoring possible constraints, randomness, and not working on anyY
     cvei(cvmat,an,col_cvmat,ymin); // check whether an is ok
     frei(an-1,&anycvY);
     frei(an-1,ad,&anycvX); 
  }
}
/************************************************************************/
/* Version : July 17 , 1996 						*/
/* Purpose : constructs xpred , outputs ypred				*/
#define COL_MISMATCH "ERROR: Number of columns in matrix does not match the number of variables\n"  
void krigpredict(KrigModel *kr,KrigArray *ka,string inputfile,string outputfile)
{
 int npred, newm; // for read xpred
 int i,pc;
 real **xpred=NULL, **ypred=NULL, **oneypred=NULL; 
 string *label=NULL, *outputlabel=NULL;
 //real **covmat=NULL;

 if (read_matrix(0,0,&xpred,&npred,&newm,inputfile,&label)!=0)
 {  
    if (npred*newm!=0) frei(npred,newm,&xpred);
    if (newm!=0)       frei(newm,&label);
    return;
 }
 if (newm!= *kr->d)
   {fprintf(ofp,COL_MISMATCH);frei(npred,newm,&xpred);frei(newm,&label);return;}

 pc = ka->p+ka->c;
 alloc(npred,pc*2,&ypred); alloc(npred,2,&oneypred);
 alloc(pc*2,&outputlabel);
 for (i=0;i< ka->p+ka->c;i++)
 {
   fprintf(ofp,"Predicting %s ... \n",kr[i].ylabel);
   sprintf(outputlabel[2*i],"Pred.%s",kr[i].ylabel);
   sprintf(outputlabel[2*i+1],"SE.%s",kr[i].ylabel);

   if (initialized(kr[i].init,"X","y","alpha","theta","beta")
    && initialized(kr[i].init,"var",NULL,NULL,NULL,NULL)  )
   { 
     ensure_init(&kr[i],"R"); ensure_init(&kr[i],"F");
     kr_cover_augmR(&kr[i]); 
     init_krigblup_anyXY(&kr[i], kr[i].X,*kr[i].n,*kr[i].d,*kr[i].r);
     krigblup_anyXY(&kr[i],kr[i].X,kr[i].y,*kr[i].n,*kr[i].d, 
       npred,xpred,oneypred);
     free_krigblup_anyXY(*kr[i].n,*kr[i].r);

     copy_matrix(npred,2,oneypred,0,ypred,2*i);

     // to test the covariances
     // cout << "npred " << npred << endl;
     //alloc(npred,npred,&covmat);
     //krigcov(kr,npred,xpred,covmat);
     //printmatr(npred,npred,covmat,"covmat:");
   }
 }
 outputmatr_cover(npred,pc*2,ypred,outputfile,outputlabel,"PredictionMatrix");
 frei(npred,newm,&xpred);frei(newm,&label); 
 frei(npred,pc*2,&ypred); frei(npred,2,&oneypred);
 frei(pc*2,&outputlabel);  
}
#undef COL_MISMATCH
/************************************************************************/
/* Version : Jan 31, 1997                                               */
/* Purpose : Calc BLUP, MSE given any X (not the one from KrigModel) 	*/
/* input   : KrigModel							*/
/*	anyX   (an,ad) X matrix						*/   
/*	anyY   (an,1) y vector 						*/   
/* 	an ,ad  dimensions of anyX					*/
/*	xpred  npred,d matrix of prediction values			*/
/* NOTE : do not change address of xpred ypred (see krigcv_anyX)	*/
/* output : ypred (npred,2) matrix of blup's and mse's			*/
void krigblup_anyXY(KrigModel *kr,real **anyX,real *anyY, int an,int ad,
     int npred,real **xpred, real **ypred)
{
  int i,n,r, counter; 
  real mse, s, rmse;
  real *vec1=NULL, *yvec=NULL;
  real *r1=NULL;
  real *temp1=NULL, *temp2=NULL;
 
  if (!mse_anyX_flag) 
    nrerror("krigmse_anyX: Code Bug: function call without its initialization");
  n = *kr->n ;  
  r = *kr->r ;
  if (ad!= *kr->d) nrerror("mse_anyX : Code Bug : ad!= *kr->d");

  alloc(r+an,&vec1); alloc(r+an,&yvec); 
  alloc(r+an,&temp1); alloc(r+an,&temp2);
  alloc(an,&r1);  /* need full r vector */   

  for (counter=0; counter<npred; counter++)
  {
     // construct first half of  vec1=(f1,r1)' yvec=(0,y)' 
     if (r==1)
     { 
        vec1[0]= 1.0; yvec[0]=0; 
     }
     else 
        nrerror("krigblup_anyXY(): regression model not dealt with, r>1 "); 
     calcr(xpred[counter],r1,anyX,kr->alpha,kr->theta,*kr->stochratio,an,ad);
     if (DEBUG && counter==0)
     {
        print_corpars(kr);
        fprintf(ofp,"an %d, ad %d\n",an,ad);
        printmatr(an,ad,anyX,"anyX");
     } 

     // construct second half of  vec1=(f1,r1)'  yvec=(0,y)' 
     for (i=0; i< an; i++){
         vec1[i+r]= r1[i]; yvec[i+r] = anyY[i];
     }
     if (DEBUG) {
        printvecr(an+r,vec1,"vec1:"); 
        printvecr(an+r,yvec,"yvec:"); 
     } 
  
     /* compute temp1 = (vec1' U) Diag(eigval)^{-1} */
     VecMultMatReal(vec1,an+r,aU,an+r,temp1); /*need for both MSE and BLUP*/
     VecEwMultVecReal(a_eigvinv,an+r,temp1); /*need for both MSE and BLUP*/

     /* compute BLUP */
     /* multiply blup  = (vec1' U) Diag(eigval)^{-1}* (yvec' U)' */
     /* multiply       = temp1'                     *  temp2     */
     VecMultMatReal(yvec,an+r,aU,an+r,temp2);
     ypred[counter][0] =VecMultVecReal(temp1,an+r,temp2);

     /* compute MSE */
     /* multiply s    = (vec1' U) Diag(eigval)^{-1}*  (vec1' U)' */
     /*               = temp1'                     *     temp2  */
     VecMultMatReal(vec1,an+r,aU,an+r,temp2); 
     s=VecMultVecReal(temp1,an+r,temp2);
     mse =  *kr->var * (1.0 -  s);
     if (mse<0.0) 
        mse = variance_check(kr,mse,counter);
     ypred[counter][1]=sqrt(mse);
     rmse = sqrt(mse);
  }
  frei(r+an,&vec1);
  frei(r+an,&temp1); frei(r+an,&temp2);
  frei(an,&r1);  
}
/************************************************************************/
/* Version : Feb 05, 1997                                               */
/* Purpose : Allocate R,U,F,eigval,eigvalinv and compute them 		*/
/*   This procedure MUST be called before the first call of krigmse_anyX*/
/* Note :  especially the computing is very time consuming and should   */
/*  not be done over and over in krigmse_anyX				*/
void init_krigblup_anyXY(KrigModel *kr, real **anyX, int an, int ad, int r)
{
  if (!mse_anyX_flag)
  {
    mse_anyX_flag=TRUE;
    alloc(an+r,an+r,&aU); 
    alloc(an,an,&aR); alloc(an,r,&aF);
    alloc(an+r,&a_eigv); alloc(an+r,&a_eigvinv);

    calcF(an,r,aF);
    calcR(anyX,kr->alpha,kr->theta,*kr->stochratio,an,ad,aR); 
    calc_augmR(an,r,aR,aF,aU,a_eigv,a_eigvinv); //calc augmR and decompose to U
  }
  else
    nrerror("init_krigmse_anyX: Code Bug : krigmse_anyX already initialized");
}
/************************************************************************/
/* Version : Feb 05, 1997                                               */
/* Purpose : Free R,U,F,eigval,eigvalinv 				*/
/*	This procedure ought to be called after the last call of 	*/
/*	krigmse_anyX, otherwise the (huge!) storage space is not freed  */ 
void free_krigblup_anyXY(int an, int r)
{
  if (mse_anyX_flag)
  {
    frei(an+r,an+r,&aU);
    frei(an,an,&aR); frei(an,r,&aF);
    frei(an+r,&a_eigv); frei(an+r,&a_eigvinv);
    mse_anyX_flag=FALSE;
  } 
  else 
    nrerror("free_krigmse_anyX: Code Bug : mse_anyX was not initialized"); 
}
/************************************************************************/
