/*****************************************************************/
/*   NELDER-MEAD SIMPLEX ALGORITHM                               */
/*****************************************************************/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "general.h"
#include "matrix.h"
#include "alloc.h"

/*******************************+++*******************************/
/*   Version:  May 14, 1996                                      */
/*                                                               */
/*   unsigned  Simplex(real (*ObjFunc)(real *x, int nDims),   */
/*                  real AbsTol, real RelTol, unsigned MaxFuncs, */
/*                  int nDims, real **p, real *y)             */
/*                                                               */
/*   Purpose:  Function minimization by the Nelder-Mead simplex  */
/*             method.                                           */
/*                                                               */
/*   Args:     ObjFunc   The objective function.                 */
/*             AbsTol    Absolute tolerance on function value    */
/*                       for convergence.                        */
/*             RelTol    Relative tolerance on function value    */
/*                       for convergence.                        */
/*             MaxFuncs  Maximum function evaluations.           */
/*             nDims     Number of dimensions.                   */
/*             p         Input:  Simplex of nDims + 1 starting   */
/*                               points;                         */
/*                       Output: Final simplex; the best point   */
/*                               is in the first row.            */
/*             y         Input:  Objective values corresponding  */
/*                               to the points in the simplex;   */
/*                       Output: Final objectives, y[0] is the   */
/*                               best.                           */
/*                                                               */
/*   Return:   Number of function evaluations.                   */
/*                                                               */
/*   Note  :   The original simplex only contains a relative     */
/*   stopping criterion. That is good in the sense that it works */
/*   independently of the actual function values. It leads       */ 
/*   to a crash in the case where f=x*x, for example, because the*/ 
/*   simplex thinks it can do (relatively) better and will never */
/*   stop, eventually deviding over 0. 				 */
/*   The absolute criterion is bad when the function values      */
/*   are close to 0, because it will lead to a too early stop    */
/*   of the optmization. In this case it may be best to take     */
/*   the absolute criterion out which can be accomplished        */
/*    by setting AbsTol= 0.0                                     */
/*                                                               */
/*****************************************************************/

/* Parameters for expansions and contractions. */
#define ALPHA 1.0
#define BETA  0.5
#define GAMMA 2.0

real *ptry = NULL;   /* Used by SimpTry for trial point. */

unsigned Simplex(real (*ObjFunc)(real *x, int nDims),
          real AbsTol, real RelTol, unsigned MaxFuncs,
          int nDims, real **p, real *y)
{
     real SimpTry(real **p, real *y, real *psum, int nDims,
          real (*ObjFunc)(real *x, int nDims), int ihi,
          unsigned *NumFuncs, real fac);
     int    i, ilo, ihi, inhi, j, NumPts;
     real      *psum=NULL;
     real      *PtrSave=NULL;
     real      ave, range, ysave, ytry;
     unsigned  NumFuncs;

     NumFuncs = 0;
     NumPts   = nDims + 1;

     alloc((int) nDims,&psum);
     alloc((int) nDims,&ptry);

     for (j = 0; j < nDims; j++)
          for (i = 0, psum[j] = 0.0; i < NumPts; i++)
               psum[j] += p[i][j];

     while (1)
     {
          /* Determine which point is highest (worst), */
          /* next highest, and lowest (best).          */
          ilo = 0;
          ihi = y[0] > y[1] ? (inhi = 1, 0) : (inhi = 0, 1);
          for (i = 0; i < NumPts; i++)
          {
               if (y[i] < y[ilo])
                    ilo=i;
               if (y[i] > y[ihi])
               {
                    inhi=ihi;
                    ihi=i;
               }
               else if (y[i] > y[inhi])
                    if (i != ihi)
                         inhi=i;
          }

          /* Stop? */
          range = y[ihi] - y[ilo]; // highest - lowest
          /* Have to be careful here to avoid overflow when */
          /* y[ilo] = y[ihi] = HUGE_VAL.                    */
          ave = 0.5 * fabs(y[ihi]) + 0.5 * fabs(y[ilo]);
          if (range <= AbsTol || range <= RelTol * ave ||
                  NumFuncs >= MaxFuncs)
               break;

          /* Extrapolate by a factor ALPHA through the face of */
          /* the simplex across from the high point.           */
          ytry = SimpTry(p, y, psum, nDims, ObjFunc, ihi,
                    &NumFuncs, -ALPHA);
          if (ytry <= y[ilo])
               /* Gives an improvement, so try an additional */
               /* extrapolation by a factor GAMMA.           */
               ytry = SimpTry(p, y, psum, nDims, ObjFunc, ihi,
                         &NumFuncs, GAMMA);
          else if (ytry >= y[inhi])
          {
               /* Worse than the second highest, so contract. */
               ysave = y[ihi];
               ytry = SimpTry(p, y, psum, nDims, ObjFunc, ihi,
                         &NumFuncs, BETA);
               if (ytry >= ysave)
               {
                    /* Can't get rid of high point.  */
                    /* Contract around th low point. */
                    for (i = 0; i < NumPts; i++)
                    {
                         if (i != ilo)
                         {
                              for (j = 0; j < nDims; j++)
                              {
                                   psum[j] = 0.5 * (p[i][j]
                                             + p[ilo][j]);
                                   p[i][j] = psum[j];
                               }
                              y[i] = (*ObjFunc)(psum, nDims);
                         }
                    }
                    NumFuncs += nDims;
                    for (j = 0; j < nDims; j++)
                         for (i = 0, psum[j] = 0.0; i < NumPts; i++)
                              psum[j] += p[i][j];
               }
          }
     }

     /* Put best point in first row of simplex, */
     /* and its objective in y[0].              */
     PtrSave = p[0];
     p[0]    = p[ilo];
     p[ilo]  = PtrSave;
     ysave   = y[0];
     y[0]    = y[ilo];
     y[ilo]  = ysave;
     frei((int) nDims, &psum);
     frei((int) nDims, &ptry);
     return(NumFuncs);
}

real SimpTry(real **p, real *y, real *psum, int nDims,
          real (*ObjFunc)(real *x, int nDims), int ihi,
          unsigned *NumFuncs, real fac)

/* Extrapolates by a factor fac across the face of the simplex */
/* from the high point, tries it, and replaces the high point  */
/* if the new point is better.                                 */

{
     int    j;
     real      fac1, fac2, ytry;

     fac1 = (1.0 - fac) / nDims;
     fac2 = fac1 - fac;
     for (j = 0; j < nDims; j++)
          ptry[j] = psum[j] * fac1 - p[ihi][j] * fac2;
     ytry = (*ObjFunc)(ptry, nDims);
     ++(*NumFuncs);

     if (ytry < y[ihi])
     {
          /* Improvement, so replace the highest point. */
          y[ihi] = ytry;
          for (j = 0; j < nDims; j++)
          {
               psum[j] += ptry[j] - p[ihi][j];
               p[ihi][j] = ptry[j];
          }
     }

     return(ytry);
}
/**************************************************************************/
/* Version : November 15, 1996                                          */
/* Purpose : minimize ObjFunc with algorithm specified in method        */
int Minimize(real (*ObjFunc)(real *x,int nDims),
          real AbsTol, real RelTol, unsigned MaxFuncs,
          int nDims, real **startx, real *starty,string method)
{
 if (strcmp(method,"simplex")==0)
   return Simplex(ObjFunc,AbsTol,RelTol,MaxFuncs,nDims,startx,starty);
 else if (strcmp(method,"powell")==0)
 {
   static real **direction=NULL;
   static int  old_d= -1;
   int niter;
   
   if (direction==NULL || old_d!=nDims)  
   { // this will crash if d has changed 
      if (nDims==0) nrerror("Minimize(Powell) : Dimension of ObjFunc is 0");
      if (direction!=NULL)
        morealloc(old_d,old_d,&direction,nDims,nDims);
      else 
        alloc(nDims,nDims,&direction);
      old_d = nDims;
      init_diag_mat(1.0,nDims,direction);
   }
   //printvecr(nDims,startx[0],"Minimizestartx");
   // only use first row of matrix x 
   niter = Powell(ObjFunc,RelTol,MaxFuncs,nDims,startx[0],starty,direction);
   //we are not freeing the space!! frei(nDims,nDims,&direction);
   return niter;
 }
 else
   nrerror("minimize: Minimization method unknown");
 return -1; //can't happen , to avoid warnings while compiling
}
/************************************************************************/
/*
real objfunc(real *x, int d) 
{
  return (x[0]*x[d-1])* (x[0]*x[d-1]);
}

void main()
{
 int it, d;
 real **x=NULL, *y=NULL; 
 
  d = 2;
  alloc(d+1,d,&x);
  alloc(d+1,&y);
  x[0][0]=5.0; x[0][1]=10.0;
  x[1][0]=3.0; x[1][1]=90.0;
  x[2][0]= -5.0; x[2][1]=10.0;
  y[0] = objfunc(x[0],d);
  y[1] = objfunc(x[1],d);
  y[2] = objfunc(x[2],d);
  
  printmatr(d+1,d,x,"xstart");
  printvecr(d+1,y,"ystart");

  it = Simplex(objfunc,0.0001,0.0001,100,2,x,y);
 
  printmatr(d+1,d,x,"xstart");
  printvecr(d+1,y,"ystart");
  printf("%d it \n",it);
}
*/


/*
real objfunc(real *x, int d) 
{
  return (x[0]*x[d-1])* (x[0]*x[d-1]);
}

void main()
{
 int it, d;
 real **x=NULL, *y=NULL; 
 
  d = 2;
  alloc(d+1,d,&x);
  alloc(d+1,&y);
  x[0][0]=5.0; x[0][1]=10.0;
  x[1][0]=3.0; x[1][1]=90.0;
  x[2][0]= -5.0; x[2][1]=10.0;
  y[0] = objfunc(x[0],d);
  y[1] = objfunc(x[1],d);
  y[2] = objfunc(x[2],d);
  
  printmatr(d+1,d,x,"xstart");
  printvecr(d+1,y,"ystart");

  it = Simplex(objfunc,0.0001,0.0001,100,2,x,y);
 
  printmatr(d+1,d,x,"xstart");
  printvecr(d+1,y,"ystart");
  printf("%d it \n",it);
}
*/
